#!/usr/bin/env python
import json
import pandas as pd
from Bio import SeqIO
from tooldistillator.interfaces import (
    ToolDistillatorResultIterator,
    logger,
)

required_metadata = {
    "hid": "Historic ID for plasmidfinder file from galaxy",
    "analysis_software_version": "plasmidfinder version",
    "reference_database_version": "plasmidfinder DB version",
    "plasmid_result_tabular_path": "plasmidfinder results in tabular format",
    "plasmid_result_tabular_hid": "plasmidfinder results hid in Galaxy",
    "genome_hit_path": (
        "fasta file with hits in genome, doesn't work for multiple input"
    ),
    "genome_hit_hid": "Historic ID for genome hit file from galaxy",
    "plasmid_hit_path": (
        "fasta file with plasmid sequences, doesn't work for multiple input"
    ),
    "plasmid_hit_hid": "Historic ID for plasmid sequence hit file from galaxy",
}


class PlasmidFinderIterator(ToolDistillatorResultIterator):
    """Plasmidfinder tool parser
    Filter results from Plasmidfinder json result
    Could also parse file for:
    Plasmid sequence hit file
    Genome sequence hit

    Args:
        ToolDistillatorResultIterator: abstract class to the parser
    """

    def __init__(self, source, metadata):
        metadata["analysis_software_name"] = "plasmidfinder"
        self.metadata = metadata
        super().__init__(source, self.metadata)
        self.tooldistillator_results["tool_has_database"] = True
        self.parse()

    def parse(self):
        """Parse plasmidfinder file"""
        self._get_plasmid_results()
        self._fill_tooldistillator_results(
            filetype="plasmid_result_tabular_path"
        )
        self._get_genome_hit_results("genome_hit_path")
        self._get_genome_hit_results("plasmid_hit_path")

    def __get_fasta_infos__(self, input: str):
        """Extract fasta informations
        Extract name, sequences, positions
        and IS information
        Returns:
            dict: sequence informations
        """
        fasta_sequences = SeqIO.parse(input, format="fasta")
        parsed_fasta_sequence = {}
        for seqname in fasta_sequences:
            sequence_infos = pd.DataFrame(seqname.description.split(" "))
            sequence_infos[[0, 1]] = sequence_infos[0].str.split(
                pat=":", n=1, expand=True
            )
            sequence_infos[[0, 1]] = sequence_infos[0].str.split(
                pat="=", n=1, expand=True
            )
            sequence_infos = sequence_infos.iloc[0:4, 0:2]
            sequence_infos = dict(
                zip(
                    sequence_infos[0],
                    sequence_infos[1],
                )
            )
            sequence_name = list(sequence_infos.keys())[0]
            sequence_infos.pop(sequence_name)
            parsed_fasta_sequence[sequence_name] = {
                "sequence_informations": sequence_infos,
                "nucleotide_sequence": seqname.seq.__str__(),
            }
        return parsed_fasta_sequence

    def item_generator(self, json_input, key):
        if isinstance(json_input, dict):
            for k, v in json_input.items():
                if k == key:
                    # self._sanitize_no_results(v)
                    yield v
                else:
                    yield from self.item_generator(v, key)
        elif isinstance(json_input, list):
            for item in json_input:
                yield from self.item_generator(item, key)

    def extract_plasmid_infos(
        self, groups=["enterobacteriales", "Gram Positive"]
    ):
        """Extract informations from plasmid json results
        Returns:
            list: a list of data
        """
        plasmid_infos = {}
        if self.stream:
            json_object = json.loads(self.stream.read())

            for key in groups:
                item = self.item_generator(json_input=json_object, key=key)
                for i in item:
                    plasmid_infos[key] = i
        return plasmid_infos

    def _sanitize_no_results(self, data, target_value="No hit found"):
        if isinstance(data, dict):
            # Iterate over the keys to avoid modif
            for key in list(data.keys()):
                if data[key] == target_value:
                    del data[key]
                else:
                    self._sanitize_no_results(data[key], target_value)
        elif isinstance(data, list):
            # Iterate over the list to avoid modifying the list while iterating
            for item in list(data):
                if item == target_value:
                    data.remove(item)
                else:
                    self._sanitize_no_results(item, target_value)

    def _get_genome_hit_results(self, filetype):
        file_path = self.metadata.get(filetype)
        self._check_input_files(file_path, optional=True)
        if self.stream:
            logger.info(f"Analysis of {filetype}")
            reader = [self.__get_fasta_infos__(file_path)]
        else:
            logger.info(f"Non informations from {filetype}")
            reader = []

        self._fill_tooldistillator_results(reader=reader, filetype=filetype)

    def _get_plasmid_results(self):
        if self.stream:
            plasmid_infos = self.extract_plasmid_infos()
            concatenated_results = self._concat_plasmid_infos(plasmid_infos)
            if concatenated_results is not None:
                concatenated_results = self._update_plasmid_infos(
                    concatenated_results
                )
                reader = self._translate_df_to_dict(concatenated_results)
                self.tooldistillator_results["results"][0]["content"] = reader

    def _fill_tooldistillator_results(self, filetype: str, reader=[]):
        """Fill the final abromics object with results

        Args:
            filetype (str): _description_
            reader (pandas.core.frame.DataFrame): trimmed pandas dataframe
        """
        hid_info = filetype.replace("path", "hid")
        report_name = filetype.replace("_path", "")
        self.tooldistillator_results["results"].append(
            {
                "name": report_name,
                "file_path": self.tooldistillator_results.pop(filetype),
                "hid": self.tooldistillator_results.pop(hid_info),
                "galaxy_id": "",
                "content": reader,
            }
        )

    def _extract_enterobacteriales_infos_to_df(
        self, plasmid_infos: dict, type="enterobacteriales"
    ):

        if plasmid_infos.get(type) not in ["No hit found", None]:

            tmp_df = pd.DataFrame.from_dict(
                plasmid_infos.get(type), orient="index"
            )
            tmp_df["inc_group"] = None
            tmp_df["plasmid_group"] = type.capitalize()
            return tmp_df
        else:
            return None

    def _extract_gram_positive_infos_to_df(
        self, plasmid_infos: dict, type="Gram Positive"
    ):
        df_list = []
        for k in plasmid_infos.get(type):
            if plasmid_infos.get(type).get(k) not in ["No hit found", None]:
                tmp_df = pd.DataFrame.from_dict(
                    plasmid_infos.get(type).get(k), orient="index"
                )
                tmp_df["inc_group"] = k
                tmp_df["plasmid_group"] = type
                df_list.append(tmp_df)

        try:
            return pd.concat(df_list)
        except ValueError:
            return None

    def _concat_plasmid_infos(self, plasmid_infos: dict):
        enterobact_results = self._extract_enterobacteriales_infos_to_df(
            plasmid_infos
        )
        gram_positive_results = self._extract_gram_positive_infos_to_df(
            plasmid_infos
        )
        if enterobact_results is not None and gram_positive_results is None:
            return enterobact_results
        elif enterobact_results is None and gram_positive_results is not None:
            return gram_positive_results
        elif enterobact_results is None and gram_positive_results is None:
            return pd.DataFrame(
                columns=[
                    "plasmid",
                    "identity",
                    "HSP_length",
                    "template_length",
                    "position_in_ref",
                    "contig_name",
                    "positions_in_contig",
                    "note",
                    "accession",
                    "coverage",
                    "hit_id",
                    "inc_group",
                    "plasmid_group",
                ]
            )
        else:
            concat_results = pd.concat(
                [enterobact_results, gram_positive_results]
            )
            return concat_results

    def _update_plasmid_infos(self, plasmid_info):
        plasmid_info["contig_name"] = (
            plasmid_info["contig_name"].str.split().str[0]
        )
        # Define a regular expression to extract all characters before ".."
        pattern_before = r"(.+)\.\."
        # Define a regular expression to extract all characters after ".."
        pattern_after = r"\.\.(.+)"
        plasmid_info["ref_start"] = plasmid_info[
            "position_in_ref"
        ].str.extract(pattern_before)
        plasmid_info["ref_end"] = plasmid_info["position_in_ref"].str.extract(
            pattern_after
        )
        plasmid_info["contig_start"] = plasmid_info[
            "positions_in_contig"
        ].str.extract(pattern_before)
        plasmid_info["contig_end"] = plasmid_info[
            "positions_in_contig"
        ].str.extract(pattern_after)
        plasmid_info.rename(columns={"note": "gene_name"}, inplace=True)
        return plasmid_info

    def _translate_df_to_dict(self, dataframe):
        return pd.DataFrame.to_dict(dataframe, orient="records")
