#!/usr/bin/env python

from Bio import SeqIO
import re
from tooldistillator.interfaces import (
    ToolDistillatorResultIterator,
)

required_metadata = {
    "hid": "Historic ID to shovill contigs file from Galaxy",
    "analysis_software_version": "shovill version number",
    "reference_database_version": "DB version",
    "contig_graph_path": "Assembly graph file",
    "contig_graph_hid": "Historic ID to assembly graph from Galaxy",
    "bam_file_path": "Binary Alignment file from shovill",
    "bam_file_hid": "Historic ID to alignment file from Galaxy",
}


class ShovillIterator(ToolDistillatorResultIterator):
    """Shovill tool parser
    Filter results from contig fasta file
    Could also keep path for:
    Alignemnt file (bam format)
    Contig graph file (gfa format)

    Args:
        ToolDistillatorResultIterator: abstract class to the parser
    """

    def __init__(self, source, metadata):
        metadata["analysis_software_name"] = "shovill"
        self.metadata = metadata
        super().__init__(source, metadata)
        self.tooldistillator_results["tool_has_database"] = False
        self.parse()

    def parse(self):
        """Assembly shovill parser
        Read contig assembly file to extract informations and sequences
        Could also add path information for
        BAM file and GFA file
        """
        if self.stream:
            parsed_fasta = self.__get_fasta_infos__()
            self.tooldistillator_results["results"][0][
               "content"] = parsed_fasta
        self.extract_metadata_information(self.metadata)

    def __get_fasta_infos__(self):
        """Extract fasta informations
        Extract name, sequences, positions
        and IS information
        Returns:
            dict: sequence informations
        """
        fasta_sequences = SeqIO.parse(self.stream, format="fasta")
        parsed_fasta_sequence = []
        for seqname in fasta_sequences:
            seq_results = {}
            seq_results["name"] = seqname.id
            seq_results["length"] = len(seqname.seq)
            seq_results["coverage"] = float(
                re.match(
                    r".*cov=([0-9.]*)",
                    seqname.description,
                ).group(1)
            )
            seq_results["sequence"] = seqname.seq.__str__()
            parsed_fasta_sequence.append(seq_results)
        return parsed_fasta_sequence
